import logging
import os
import pandas as pd


from openfisca_survey_manager.paths import openfisca_survey_manager_location

log = logging.getLogger(__name__)


legislation_directory = os.path.join(
    openfisca_survey_manager_location,
    'openfisca_survey_manager',
    'assets',
    )


sub_levels = ['divisions', 'groupes', 'classes', 'sous_classes', 'postes']
divisions = ['0{}'.format(i) for i in range(1, 10)] + ['11', '12']


def build_coicop_level_nomenclature(level, year = 2016, keep_code = False, to_csv = False):
    assert level in sub_levels
    log.debug("Reading nomenclature coicop {} source data for level {}".format(year, level))
    try:
        if year == 1998:
            data_frame = pd.read_csv(
                os.path.join(legislation_directory, 'COICOP/1998/nomenclature_coicop1998_source_by_{}.csv'.format(level)),
                sep = ';',
                header = None,
            )
        if year == 2016:
            data_frame = pd.read_excel(
                os.path.join(legislation_directory, 'COICOP/2016/nomenclature_coicop2016_source_by_{}.xls'.format(level)),
                header = None,)

    except Exception as e:
        log.info("Error when reading nomenclature coicop source data for level {}".format(level))
        raise e

    data_frame.reset_index(inplace = True)
    data_frame.rename(columns = {0: 'code_coicop', 1: 'label_{}'.format(level[:-1])}, inplace = True)
    data_frame = data_frame.iloc[2:].copy()
    if year == 2016:
        data_frame['code_coicop'] = data_frame['code_coicop'].apply(lambda x: x[1:])

    index, stop = 0, False
    for sub_level in sub_levels:
        if stop:
            continue
        if sub_level == 'divisions':
            data_frame[sub_level] = data_frame['code_coicop'].str[index:index + 2].astype(int)
            index = index + 3
        else:
            data_frame[sub_level] = data_frame['code_coicop'].str[index:index + 1].astype(int)
            index = index + 2

        if level == sub_level:
            stop = True

    if keep_code or level == 'postes':
        data_frame['code_coicop'] = data_frame['code_coicop'].str.lstrip("0")
    else:
        del data_frame['code_coicop']

    data_frame.reset_index(inplace = True, drop = True)
    if to_csv:
        data_frame.to_csv(
            os.path.join(legislation_directory, 'nomenclature_coicop{}_by_{}.csv'.format(year, level)),
            )

    return data_frame


def build_raw_coicop_nomenclature(year = 2016):
    """Builds raw COICOP nomenclature from ecoicop levels"""
    coicop_nomenclature = None

    for index in range(len(sub_levels) - 1):
        level = sub_levels[index]
        next_level = sub_levels[index + 1]
        on = sub_levels[: index + 1]

        df_left = coicop_nomenclature if coicop_nomenclature is not None else build_coicop_level_nomenclature(level, year)
        df_right = build_coicop_level_nomenclature(next_level, year)

        # Drop any residual 'index' columns to avoid merge conflicts
        for df in (df_left, df_right):
            if "index" in df.columns:
                df.drop(columns=["index"], inplace=True)

        coicop_nomenclature = pd.merge(
            df_left,
            df_right,
            on=on,
            how="inner",
            validate="one_to_many",  # safety check
        )

    # Reorder and select relevant columns
    coicop_nomenclature = coicop_nomenclature[
        ["code_coicop"]
        + [f"label_{sub_level[:-1]}" for sub_level in sub_levels]
        + sub_levels
    ].copy()

    return coicop_nomenclature[
        ["label_division", "label_groupe", "label_classe",
         "label_sous_classe", "label_poste", "code_coicop"]
    ].copy()


if __name__ == "__main__":
    import sys
    logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    raw_coicop_nomenclature = build_raw_coicop_nomenclature()
    log.info(raw_coicop_nomenclature)



"""Tests for the survey scenario functionality in OpenFisca Survey Manager."""

import shutil
import logging
import os
import pytest
from typing import Dict, Any, List, Optional, Callable


from openfisca_core import periods
from openfisca_core.tools import assert_near


from openfisca_survey_manager.paths import (
    openfisca_survey_manager_location,
    default_config_files_directory,
)
from openfisca_survey_manager.input_dataframe_generator import (
    make_input_dataframe_by_entity,
    random_data_generator,
    randomly_init_variable,
)
from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario
from openfisca_survey_manager.scenarios.reform_scenario import ReformScenario
from openfisca_survey_manager.tests import tax_benefit_system


log = logging.getLogger(__name__)


def create_randomly_initialized_survey_scenario(
    nb_persons: int = 10,
    nb_groups: int = 5,
    salary_max_value: float = 50000,
    rent_max_value: float = 1000,
    collection: Optional[str] = "test_random_generator",
    use_marginal_tax_rate: bool = False,
    reform: Optional[Callable] = None,
) -> AbstractSurveyScenario:
    """
    Create a randomly initialized survey scenario.

    Args:
        nb_persons (int): Number of persons
        nb_groups (int): Number of groups
        salary_max_value (float): Maximum salary value
        rent_max_value (float): Maximum rent value
        collection (Optional[str]): Collection name
        use_marginal_tax_rate (bool): Use marginal tax rate
        reform (Optional[Callable]): Reform function

    Returns:
        AbstractSurveyScenario: Initialized survey scenario
    """
    if collection is not None:
        return create_randomly_initialized_survey_scenario_from_table(
            nb_persons,
            nb_groups,
            salary_max_value,
            rent_max_value,
            collection,
            use_marginal_tax_rate,
            reform=reform,
        )
    else:
        return create_randomly_initialized_survey_scenario_from_data_frame(
            nb_persons,
            nb_groups,
            salary_max_value,
            rent_max_value,
            use_marginal_tax_rate,
            reform=reform,
        )


def create_randomly_initialized_survey_scenario_from_table(
    nb_persons: int,
    nb_groups: int,
    salary_max_value: float,
    rent_max_value: float,
    collection: str,
    use_marginal_tax_rate: bool,
    reform: Optional[Callable] = None,
) -> AbstractSurveyScenario:
    """
    Create a randomly initialized survey scenario from a table.

    Args:
        nb_persons (int): Number of persons
        nb_groups (int): Number of groups
        salary_max_value (float): Maximum salary value
        rent_max_value (float): Maximum rent value
        collection (str): Collection name
        use_marginal_tax_rate (bool): Use marginal tax rate
        reform (Optional[Callable]): Reform function

    Returns:
        AbstractSurveyScenario: Initialized survey scenario
    """
    variable_generators_by_period = {
        periods.period("2017-01"): [
            {
                "variable": "salary",
                "max_value": salary_max_value,
            },
            {
                "variable": "rent",
                "max_value": rent_max_value,
            },
            {
                "variable": "household_weight",
                "max_value": 100,
            },
        ],
        periods.period("2018-01"): [
            {
                "variable": "salary",
                "max_value": salary_max_value,
            },
        ],
    }
    table_by_entity_by_period = random_data_generator(
        tax_benefit_system,
        nb_persons,
        nb_groups,
        variable_generators_by_period,
        collection,
    )
    if reform is None:
        survey_scenario = AbstractSurveyScenario()
        survey_scenario.set_tax_benefit_systems(dict(baseline=tax_benefit_system))
    else:
        survey_scenario = ReformScenario()
        survey_scenario.set_tax_benefit_systems(
            dict(
                reform=reform(tax_benefit_system),
                baseline=tax_benefit_system,
            )
        )

    survey_scenario.used_as_input_variables = [
        "salary",
        "rent",
        "housing_occupancy_status",
        "household_weight",
    ]
    survey_scenario.period = 2017
    survey_scenario.collection = collection
    data = {
        "survey": "input",
        "input_data_table_by_entity_by_period": table_by_entity_by_period,
        "config_files_directory": default_config_files_directory,
    }
    if use_marginal_tax_rate:
        survey_scenario.varying_variable = "salary"

    survey_scenario.weight_variable_by_entity = {
        "person": "person_weight",
        "household": "household_weight",
    }
    survey_scenario.init_from_data(
        data=data, use_marginal_tax_rate=use_marginal_tax_rate
    )
    return survey_scenario


def create_randomly_initialized_survey_scenario_from_data_frame(
    nb_persons: int,
    nb_groups: int,
    salary_max_value: float,
    rent_max_value: float,
    use_marginal_tax_rate: bool = False,
    reform: Optional[Callable] = None,
) -> AbstractSurveyScenario:
    """
        Create a randomly initialized survey scenario from a data frame.

    Args:
        nb_persons (int): Number of persons
        nb_groups (int): Number of groups
        salary_max_value (float): Maximum salary value
        rent_max_value (float): Maximum rent value
        use_marginal_tax_rate (bool): Use marginal tax rate
        reform (Optional[Callable]): Reform function

    Returns:
        AbstractSurveyScenario: Initialized survey scenario
    """
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value
    )
    for entity in input_data_frame_by_entity.keys():
        if entity == "person":
            input_data_frame_by_entity[entity]["household_id_ind"] = (
                input_data_frame_by_entity[entity]["household_id"]
            )
        if entity == "household":
            input_data_frame_by_entity[entity]["household_id"] = (
                input_data_frame_by_entity[entity].index
            )

    weight_variable_by_entity = {
        "person": "person_weight",
        "household": "household_weight",
    }
    if reform is None:
        survey_scenario = AbstractSurveyScenario()
        survey_scenario.set_tax_benefit_systems(dict(baseline=tax_benefit_system))
    else:
        survey_scenario = ReformScenario()
        survey_scenario.set_tax_benefit_systems(
            dict(
                reform=reform(tax_benefit_system),
                baseline=tax_benefit_system,
            )
        )
    survey_scenario.period = 2017
    survey_scenario.used_as_input_variables = [
        "salary",
        "rent",
        "household_weight",
        "household_id",
        "household_id_ind",
    ]
    period = periods.period("2017-01")

    data = {
        "input_data_frame_by_entity_by_period": {period: input_data_frame_by_entity},
        "config_files_directory": default_config_files_directory,
    }
    survey_scenario.set_weight_variable_by_entity(weight_variable_by_entity)
    assert survey_scenario.weight_variable_by_entity == weight_variable_by_entity
    survey_scenario.init_from_data(data=data)
    if isinstance(survey_scenario, ReformScenario):
        assert (
            survey_scenario.calculate_series(
                "household_weight", period, use_baseline=True
            )
            != 0
        ).all()
    else:
        for simulation_name, simulation in survey_scenario.simulations.items():
            assert simulation.weight_variable_by_entity == weight_variable_by_entity, (
                f"{simulation_name} weight_variable_by_entity does not match {weight_variable_by_entity}"
            )
            assert (
                survey_scenario.calculate_series(
                    "household_weight", period, simulation=simulation_name
                )
                != 0
            ).all()
    return survey_scenario


def generate_input_input_dataframe_by_entity(
    nb_persons: int, nb_groups: int, salary_max_value: float, rent_max_value: float
) -> Dict[str, Any]:
    """
    Generate input dataframe by entity with randomly initialized variables.

    Args:
        nb_persons (int): Number of persons
        nb_groups (int): Number of groups
        salary_max_value (float): Maximum salary value
        rent_max_value (float): Maximum rent value

    Returns:
        Dict[str, Any]: Input dataframe by entity
    """
    input_dataframe_by_entity = make_input_dataframe_by_entity(
        tax_benefit_system, nb_persons, nb_groups
    )
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        "salary",
        max_value=salary_max_value,
        condition="household_role == 'first_parent'",
    )
    randomly_init_variable(
        tax_benefit_system, input_dataframe_by_entity, "rent", max_value=rent_max_value
    )
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        "household_weight",
        max_value=100,
    )
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        "housing_occupancy_status",
        max_value=4,
    )
    return input_dataframe_by_entity


def test_input_dataframe_generator(
    nb_persons: int = 10,
    nb_groups: int = 5,
    salary_max_value: float = 50000,
    rent_max_value: float = 1000,
) -> None:
    """Test the input dataframe generator function."""
    input_dataframe_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value
    )
    assert (
        input_dataframe_by_entity["person"]["household_role"] == "first_parent"
    ).sum() == 5
    assert (
        input_dataframe_by_entity["person"].loc[
            input_dataframe_by_entity["person"]["household_role"] != "first_parent",
            "salary",
        ]
        == 0
    ).all()

    assert (
        input_dataframe_by_entity["person"].loc[
            input_dataframe_by_entity["person"]["household_role"] == "first_parent",
            "salary",
        ]
        > 0
    ).all()
    assert (
        input_dataframe_by_entity["person"].loc[
            input_dataframe_by_entity["person"]["household_role"] == "first_parent",
            "salary",
        ]
        <= salary_max_value
    ).all()

    assert (input_dataframe_by_entity["household"]["rent"] > 0).all()
    assert (input_dataframe_by_entity["household"]["rent"] < rent_max_value).all()
    assert (
        input_dataframe_by_entity["household"]["housing_occupancy_status"]
        .isin([0, 1, 2, 3])
        .all()
    )


def test_init_from_data(
    nb_persons: int = 10,
    nb_groups: int = 5,
    salary_max_value: float = 50000,
    rent_max_value: float = 1000,
) -> None:
    """
    Test the initialization of data in the survey scenario.

    Args:
        nb_persons: Number of persons to generate in the test data.
        nb_groups: Number of household groups to generate.
        salary_max_value: Maximum value for randomly generated salaries.
        rent_max_value: Maximum value for randomly generated rents.
    """
    # We check that the attribute `used_as_input_variables` matches the list of variables
    # that are used in the calculation of the simulations. Other variables are not used in the calculation,
    # being present in the input dataset but not in the output dataset (the simulation dataset)

    # Set up test : the minimum necessary data to perform an `init_from_data`
    survey_scenario = AbstractSurveyScenario()
    assert survey_scenario.simulations is None
    # Generate some data and its period
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value
    )
    period = periods.period("2017-01")
    # Creating a data object associated to its period, and we give it a name
    data_in = {
        "input_data_frame_by_entity_by_period": {period: input_data_frame_by_entity}
    }
    # data_in = copy.deepcopy(data_in) # Pour comparer avec la sortie de `init_from_data`
    table_ind = input_data_frame_by_entity["person"].copy(deep=True)
    table_men = input_data_frame_by_entity["household"].copy(deep=True)
    # print(table_ind)

    # We must add a TBS to the scenario to indicate what are the entities
    survey_scenario.set_tax_benefit_systems(dict(baseline=tax_benefit_system))
    assert len(survey_scenario.tax_benefit_systems) == 1
    assert list(survey_scenario.tax_benefit_systems.keys()) == ["baseline"]
    assert survey_scenario.simulations is None
    # We must add the `used_as_input_variables` even though they don't seem necessary
    survey_scenario.used_as_input_variables = ["salary", "rent", "household_weight"]
    # We must add the year to initiate a .new_simulation
    survey_scenario.period = 2017
    # Then we can input the data+period dict inside the scenario
    survey_scenario.init_from_data(data=data_in)
    assert len(survey_scenario.simulations) == 1
    # We are looking for the dataframes inside the survey_scenario
    all_var = list(set(list(table_ind.columns) + list(table_men.columns)))
    # print('Variables', all_var)
    data_out = survey_scenario.create_data_frame_by_entity(
        variables=all_var, period=period, merge=False
    )

    # 1 - Has the data object changed ? We only compare variables because Id's and others are lost in the process
    for cols in table_ind:
        if cols in data_out["person"]:
            pass
        else:
            print("Columns lost in person table: ", cols)  # noqa T201
    assert data_out["person"]["salary"].equals(table_ind["salary"])

    for cols in table_men:
        if cols in data_out["household"]:
            pass
        else:
            print("Columns lost in household table: ", cols)  # noqa T201
    assert data_out["household"]["rent"].equals(table_men["rent"])


def test_survey_scenario_input_dataframe_import(
    nb_persons: int = 10,
    nb_groups: int = 5,
    salary_max_value: float = 50000,
    rent_max_value: float = 1000,
) -> None:
    """
    Test the import of input dataframes into a survey scenario.

    Args:
        nb_persons: Number of persons to generate.
        nb_groups: Number of household groups.
        salary_max_value: Maximum salary value.
        rent_max_value: Maximum rent value.
    """
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value
    )
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline=tax_benefit_system))
    survey_scenario.period = 2017
    survey_scenario.used_as_input_variables = ["salary", "rent"]
    period = periods.period("2017-01")
    data = {
        "input_data_frame_by_entity_by_period": {period: input_data_frame_by_entity},
        "config_files_directory": default_config_files_directory,
    }
    survey_scenario.init_from_data(data=data)

    simulation = survey_scenario.simulations["baseline"]
    assert (
        simulation.calculate("salary", period)
        == input_data_frame_by_entity["person"]["salary"]
    ).all()
    assert (
        simulation.calculate("rent", period)
        == input_data_frame_by_entity["household"]["rent"]
    ).all()


def test_survey_scenario_input_dataframe_import_scrambled_ids(
    nb_persons: int = 10,
    nb_groups: int = 5,
    salary_max_value: float = 50000,
    rent_max_value: float = 1000,
) -> None:
    """
    Test survey scenario input dataframe import with scrambled IDs.

    Args:
        nb_persons: Number of persons to generate.
        nb_groups: Number of household groups.
        salary_max_value: Maximum salary value.
        rent_max_value: Maximum rent value.
    """
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value
    )  # Un dataframe d'exemple que l'on injecte
    input_data_frame_by_entity["person"]["household_id"] = (
        4 - input_data_frame_by_entity["person"]["household_id"]
    )
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline=tax_benefit_system))
    survey_scenario.period = 2017
    survey_scenario.used_as_input_variables = ["salary", "rent"]
    period = periods.period("2017-01")
    data = {
        "input_data_frame_by_entity_by_period": {period: input_data_frame_by_entity},
        "config_files_directory": default_config_files_directory,
    }
    survey_scenario.init_from_data(data=data)
    simulation = survey_scenario.simulations["baseline"]
    period = periods.period("2017-01")
    assert (
        simulation.calculate("salary", period)
        == input_data_frame_by_entity["person"]["salary"]
    ).all()
    assert (
        simulation.calculate("rent", period)
        == input_data_frame_by_entity["household"]["rent"]
    ).all()


def test_dump_survey_scenario() -> None:
    """Test the dump and restore functionality of survey scenarios."""
    survey_scenario = create_randomly_initialized_survey_scenario()
    directory = os.path.join(
        openfisca_survey_manager_location,
        "openfisca_survey_manager",
        "tests",
        "data_files",
        "dump",
    )
    if os.path.exists(directory):
        shutil.rmtree(directory)

    survey_scenario.dump_simulations(directory=directory)
    period = "2017-01"
    df = survey_scenario.create_data_frame_by_entity(
        variables=["salary", "rent"], period=period
    )
    household = df["household"]
    person = df["person"]
    assert not household.empty
    assert not person.empty
    del survey_scenario
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline=tax_benefit_system))
    survey_scenario.used_as_input_variables = ["salary", "rent"]
    survey_scenario.period = 2017
    survey_scenario.restore_simulations(directory=directory)
    df2 = survey_scenario.create_data_frame_by_entity(
        variables=["salary", "rent"], period="2017-01"
    )

    assert (df2["household"] == household).all().all()
    assert (df2["person"] == person).all().all()


@pytest.mark.order(
    before="test_add_survey_to_collection.py::test_add_survey_to_collection"
)
def test_inflate() -> None:
    """Test the inflate method of the survey scenario."""
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    inflator = 2.42
    inflator_by_variable = {"rent": inflator}

    rent_before_inflate = survey_scenario.compute_aggregate("rent", period=period)
    survey_scenario.inflate(inflator_by_variable=inflator_by_variable, period=period)
    rent_after_inflate = survey_scenario.compute_aggregate("rent", period=period)
    assert_near(
        rent_after_inflate,
        inflator * rent_before_inflate,
        relative_error_margin=1e-6,
        message="Failing inflate with inflator_by_variable: rent_after_inflate = {} != {} = rent_before_inflate ({}) x inflator ({})".format(
            rent_after_inflate,
            rent_before_inflate * inflator,
            rent_before_inflate,
            inflator,
        ),
    )

    target = 3e5
    target_by_variable = {"salary": target}
    salary_before_inflate = survey_scenario.compute_aggregate("salary", period=period)
    survey_scenario.inflate(target_by_variable=target_by_variable, period=period)

    salary_after_inflate = survey_scenario.compute_aggregate("salary", period=period)
    assert_near(
        salary_after_inflate,
        target,
        relative_error_margin=1e-6,
        message="Failing inflate with inflator_by_variable: salary_after_inflate = {} != {} = target (salary_before_inflate = {})\n".format(
            salary_after_inflate,
            target,
            salary_before_inflate,
        ),
    )


@pytest.mark.order(
    before="test_add_survey_to_collection.py::test_add_survey_to_collection"
)
def test_compute_pivot_table() -> None:
    """Test the compute_pivot_table method of the survey scenario."""
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    pivot_table = survey_scenario.compute_pivot_table(
        columns=["age"], values=["salary"], period=period, simulation="baseline"
    )

    assert pivot_table.index == "salary"
    assert pivot_table.values.round() == 21748

    del survey_scenario.weight_variable_by_entity
    survey_scenario.set_weight_variable_by_entity()
    pivot_table = survey_scenario.compute_pivot_table(
        columns=["age"], values=["salary"], period=period, simulation="baseline"
    )

    assert pivot_table.values.round() == 13570.0


def test_compute_quantile() -> List[float]:
    """Test the compute_quantiles method of the survey scenario."""
    survey_scenario = create_randomly_initialized_survey_scenario()
    period = "2017-01"
    quintiles = survey_scenario.compute_quantiles(
        variable="salary",
        nquantiles=5,
        period=period,
        weighted=False,
        simulation="baseline",
    )
    return quintiles


if __name__ == "__main__":
    import sys

    log = logging.getLogger(__name__)
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
    quintiles = test_compute_quantile()
    # pivot_table = test_compute_pivot_table()
    # test_inflate()
    # test_create_data_frame_by_entity()



import logging

import numpy
from numpy import logical_not
import pandas as pd

from openfisca_core.model_api import Enum
from openfisca_survey_manager.calmar import calmar


log = logging.getLogger(__name__)


class Calibration(object):
    """An object to calibrate survey data of a SurveyScenario."""

    filter_by = None
    initial_entity_count = None
    _initial_weight_name = None
    initial_weight_by_entity = dict()
    target_margins = dict()
    margins_by_variable = dict()
    parameters = {
        'use_proportions': True,
        'initial_weight': None,
        'method': None,  # 'linear', 'raking ratio', 'logit', 'hyperbolic sinus'
        'up': None,
        'invlo': None,
        'alpha': None,
        'id_variable': None,
        'id_variable_link': None,
        }
    period = None
    simulation = None
    target_entity_count = None
    other_entity_count = None
    target_entity = None
    weight_name = None
    entities = None

    def __init__(self, simulation, target_margins, period, target_entity_count = None, other_entity_count = None, parameters = None,
            filter_by = None, entity = None):
        target_entity = entity
        self.parameters = parameters
        self.period = period
        self.simulation = simulation
        if target_margins:
            margin_variables = list(target_margins.keys())
        else:
            margin_variables = []

        variable_instance_by_variable_name = simulation.tax_benefit_system.variables
        entities = set(
            variable_instance_by_variable_name[variable].entity.key
            for variable in margin_variables
            )
        if entity is not None:
            entities.add(entity)
        self.entities = list(entities)

        if len(entities) == 0:
            assert target_entity_count != 0
            assert target_entity in [
                entity.key
                for entity in simulation.tax_benefit_system.entities
                ]
        elif len(entities) == 2:
            assert "id_variable" in parameters.keys() and parameters["id_variable"] is not None, "With two entities involved, an id variable of the largest entity is needed"
            assert "id_variable_link" in parameters.keys() and parameters["id_variable_link"] is not None, "With two entities involved, an id variable linking entity is needed"
            entity_id_variable = variable_instance_by_variable_name[parameters["id_variable"]].entity.key
            entity_id_variable_link = variable_instance_by_variable_name[parameters["id_variable_link"]].entity.key
            assert entity_id_variable in entities, "'id_variable' do not correspond to a calibrating variable entity"
            assert entity_id_variable_link in entities, "'id_variable' do not correspond to a calibrating variable entity"
            assert entity_id_variable != entity_id_variable_link, "'id_variable_link' must associate a smaller entity to the id of the greater, 'id_variable'"
            id_variable = simulation.calculate(parameters["id_variable"], period)
            id_variable_link = simulation.calculate(parameters["id_variable_link"], period)
            assert numpy.unique(id_variable_link).sort() == numpy.unique(id_variable).sort(), "There is no inclusion of one entity in the other"
            assert len(id_variable) < len(id_variable_link), "{} seems to be included in {}, not the opposite. Try reverse 'id_variable' and 'id_variable_link'".format(entity_id_variable_link, entity_id_variable)
            target_entity = entity_id_variable
        elif len(entities) > 2:
            raise NotImplementedError("Cannot hande multiple entites")
        else:
            target_entity = list(entities)[0]
            if "id_variable" in parameters.keys():
                assert variable_instance_by_variable_name[parameters["id_variable"]].entity.key == target_entity, "'id_variable' isn't the id of the entity targeted by the calibration variables"

        assert simulation.weight_variable_by_entity is not None
        if isinstance(simulation.weight_variable_by_entity, dict):
            weight_variable_name = simulation.weight_variable_by_entity[target_entity]
        else:
            weight_variable_name = simulation.weight_variable_by_entity.get(entity)
        self.weight_name = weight_name = weight_variable_name

        self.target_entity = target_entity
        period = self.period

        if filter_by:
            self.filter_by = simulation.calculate(filter_by, period = period)
        else:
            self.filter_by = numpy.array(1.0)

        assert weight_name is not None, "A calibration needs a weight variable name to act on"
        self._initial_weight_name = weight_name + "_ini"
        self.initial_weight = initial_weight = simulation.calculate(weight_name, period = period)

        self.initial_entity_count = sum(initial_weight * self.filter_by)
        self.target_entity_count = target_entity_count
        self.other_entity_count = other_entity_count

        self.weight = initial_weight.copy()

        # TODO does not work
        for entity, weight_variable in simulation.weight_variable_by_entity.items():
            self.initial_weight_by_entity[entity] = simulation.calculate(weight_variable, period = period)

        if target_margins:
            for variable, target in target_margins.items():
                self.set_target_margin(variable, target)

    def _build_calmar_data(self) -> dict:
        """Build the data dictionnary used as calmar input argument.

        Returns:
            dict containing one pd.DataFrame by entity: Data used by calmar, and the name of the aggregating entity
        """
        # Select only filtered entities
        assert self._initial_weight_name is not None
        data = dict()
        for entity in self.entities:
            data[entity] = pd.DataFrame()
        data[self.target_entity][self._initial_weight_name] = self.initial_weight * self.filter_by
        period = self.period
        for variable in self.margins_by_variable:
            assert variable in self.simulation.tax_benefit_system.variables
            data[self.simulation.tax_benefit_system.variables[variable].entity.key][variable] = self.simulation.adaptative_calculate_variable(variable, period = period)

        if len(self.entities) == 2:
            for entity in self.entities:
                if entity == self.target_entity:
                    data[entity]['id_variable'] = self.simulation.adaptative_calculate_variable(self.parameters["id_variable"], period = period)
                else:
                    data[entity][self.simulation.weight_variable_by_entity[entity]] = self.initial_weight_by_entity[entity]
                    data[entity]['id_variable'] = self.simulation.adaptative_calculate_variable(self.parameters["id_variable_link"], period = period)
        data['target_entity_name'] = self.target_entity

        return data

    def calibrate(self, inplace = False):
        """Apply the calibrations by updating weights and margins.

        Args:
            inplace (bool, optional): Whether to return the calibrated or to setthem inplace. Defaults to False.

        Returns:
            numpy.array: calibrated weights
        """
        assert self.margins_by_variable is not None, "Margins by variable should be set"
        margins_by_variable = self.margins_by_variable
        parameters = self.get_parameters()

        if margins_by_variable is not None:
            simple_margins_by_variable = dict([
                (variable, margins_by_type['target'])
                for variable, margins_by_type in margins_by_variable.items()])
        else:
            simple_margins_by_variable = dict()

        if self.target_entity_count is not None:
            simple_margins_by_variable['total_population'] = self.target_entity_count
        if self.other_entity_count is not None:
            simple_margins_by_variable['total_population_smaller_entity'] = self.other_entity_count

        self._update_weights(simple_margins_by_variable, parameters = parameters)
        self._update_margins()
        if inplace:
            self.set_calibrated_weights()
            return

        return self.weight

    def get_parameters(self) -> dict:
        """Get the parameters.

        Returns:
            dict: Parameters
        """
        p = {}
        p['method'] = self.parameters.get('method', 'linear')
        if self.parameters.get('invlo') is not None:
            p['lo'] = 1 / self.parameters.get('invlo')
        p['up'] = self.parameters.get('up')
        if p['method'] == 'logit':
            assert self.parameters.get('invlo') is not None and self.parameters.get('up') is not None
            p['lo'] = 1 / self.parameters.get('invlo')
            p['up'] = self.parameters.get('up')
        if p['method'] == 'hyperbolic sinus':
            assert self.parameters.get('alpha') is not None
            p['alpha'] = self.parameters.get('alpha')
        p['use_proportions'] = self.parameters.get('use_proportions', True)
        p['initial_weight'] = self.weight_name + ""
        return p

    def set_target_margin(self, variable, target):
        """Set variable target margin.

        Args:
          variable: Target variable
          target: Target value
        """
        simulation = self.simulation
        period = self.period
        assert variable in simulation.tax_benefit_system.variables
        variable_instance = simulation.tax_benefit_system.variables[variable]

        filter_by = self.filter_by
        target_by_category = None
        categorical_variable = (
            (variable_instance.value_type in [bool, Enum])
            or (variable_instance.unit == 'years')
            or (variable_instance.unit == 'months')
            )
        if categorical_variable:
            value = simulation.calculate(variable, period = period)
            filtered_value = value if (filter_by == numpy.array(1.0) or all(filter_by)) else value[filter_by.astype(bool)]
            categories = numpy.sort(numpy.unique(filtered_value))
            target_by_category = dict(zip(categories, target))

        if not self.margins_by_variable:
            self.margins_by_variable = dict()
        if variable not in self.margins_by_variable:
            self.margins_by_variable[variable] = dict()
        self.margins_by_variable[variable]['target'] = target_by_category or target
        self._update_margins()

    def reset(self):
        """Reset the calibration to its initial state."""
        simulation = self.simulation
        simulation.delete_arrays(self.weight_name, self.period)
        simulation.set_input(self.weight_name, self.period, numpy.array(self.initial_weight))

    def set_calibrated_weights(self):
        """Modify the weights to use the calibrated weights."""
        period = self.period
        simulation = self.simulation
        simulation.set_input(self.weight_name, period, self.weight)
        for weight_name in simulation.weight_variable_by_entity.values():
            weight_variable = simulation.tax_benefit_system.variables[weight_name]
            if weight_name == self.weight_name:
                weight_variable.unit = "base_weight"  # The weight variable is flagged as the one that have changed
                if weight_variable.formulas:
                    weight_variable.formulas = []  # The weight variable becomes an input variable after it changes with calibration
            # Delete other entites already computed weigths
            # to ensure that this weights a recomputed if they derive from
            # the calibrated weight variable
            if weight_variable.formulas:
                simulation.delete_arrays(weight_variable.name, period)

    def summary(self):
        """Summarize margins."""
        margins_df = pd.DataFrame.from_dict(self.margins_by_variable).T
        margins_df.loc['entity_count', 'actual'] = (self.weight * self.filter_by).sum()
        margins_df.loc['entity_count', 'initial'] = (self.initial_weight * self.filter_by).sum()
        margins_df.loc['entity_count', 'target'] = self.target_entity_count
        return margins_df

    def _update_margins(self):
        """Update margins."""
        for variable in self.margins_by_variable:
            simulation = self.simulation
            period = self.period
            target_entity = self.target_entity

            # These are the varying weights
            weight = self.weight
            filter_by = self.filter_by
            initial_weight = self.initial_weight

            value = simulation.adaptative_calculate_variable(variable, period = period)
            weight_variable = simulation.weight_variable_by_entity[target_entity]

            if len(self.entities) == 2 and simulation.tax_benefit_system.variables[variable].entity.key != self.target_entity:
                value_df = pd.DataFrame(value)
                id_variable = self.parameters["id_variable_link"]
                value_df[id_variable] = simulation.adaptative_calculate_variable(id_variable, period = period)
                value = value_df.groupby(id_variable).sum().to_numpy()

            if filter_by != 1:
                if weight_variable != self.weight_name:
                    NotImplementedError("No filtering possible so far when target varaible is not on the same entity as varying weights")

                weight = weight[filter_by]
                initial_weight = initial_weight[filter_by]
                value = value[filter_by]

            margin_items = [
                ('actual', weight),
                ('initial', initial_weight),
                ]

            variable_instance = simulation.tax_benefit_system.get_variable(variable)
            assert variable_instance is not None
            if variable_instance.value_type in [bool, Enum]:
                margin_items.append(('category', value))
                margins_data_frame = pd.DataFrame.from_items(margin_items)
                margins_data_frame = margins_data_frame.groupby('category', sort = True).sum()
                margin_by_type = margins_data_frame.to_dict()
            else:
                margin_by_type = dict(
                    actual = (weight * value).sum(),
                    initial = (initial_weight * value).sum(),
                    )
            self.margins_by_variable[variable].update(margin_by_type)

    def _update_weights(self, margins, parameters = None):
        """Run calmar, stores new weights and returns adjusted margins.

        Args:
          margins: margins
          parameters:  Parameters (Default value = {})

        Returns:
            dict: Updated margins

        """
        if parameters is None:
            parameters = dict()

        target_entity = self.target_entity
        weight_variable = self.simulation.weight_variable_by_entity[target_entity]

        if self.weight_name != weight_variable:
            raise NotImplementedError("Calmar needs to be adapted. Consider using a projected target on the entity with changing weights")

        data = self._build_calmar_data()

        assert self._initial_weight_name is not None
        parameters['initial_weight'] = self._initial_weight_name
        if self.target_entity_count:
            margins["total_population"] = self.target_entity_count
        if self.other_entity_count:
            margins["total_population_smaller_entity"] = self.other_entity_count

        val_pondfin, lambdasol, updated_margins = calmar(
            data, margins, **parameters)
        # Updating only after filtering weights
        self.weight = val_pondfin * self.filter_by + self.weight * (logical_not(self.filter_by))

        return updated_margins



"""CALMAR: Calibrates weights to satisfy margins constraints."""

import logging
import operator
import pandas as pd

from numpy import exp, ones, zeros, unique, array, dot, float64, sqrt
from numpy import log as ln


log = logging.getLogger(__name__)


def linear(u):
    """

    Args:
      u:

    Returns:

    """
    return 1 + u


def linear_prime(u):
    """

    Args:
      u:

    Returns:

    """
    return ones(u.shape, dtype = float)


def raking_ratio(u):
    """

    Args:
      u:

    Returns:

    """
    return exp(u)


def raking_ratio_prime(u):
    """

    Args:
      u:

    Returns:

    """
    return exp(u)


def logit(u, low, up):
    """

    Args:
      u:
      low:
      up:

    Returns:

    """
    a = (up - low) / ((1 - low) * (up - 1))
    return (low * (up - 1) + up * (1 - low) * exp(a * u)) / (up - 1 + (1 - low) * exp(a * u))


def logit_prime(u, low, up):
    """

    Args:
      u:
      low:
      up:

    Returns:

    """
    a = (up - low) / ((1 - low) * (up - 1))
    return (
        (a * up * (1 - low) * exp(a * u)) * (up - 1 + (1 - low) * exp(a * u))
        - (low * (up - 1) + up * (1 - low) * exp(a * u)) * (1 - low) * a * exp(a * u)
        ) / (up - 1 + (1 - low) * exp(a * u)) ** 2


def hyperbolic_sinus(u, alpha):
    logarithm = ln(2 * alpha * u + sqrt(4 * (alpha ** 2) * (u ** 2) + 1))
    return 0.5 * (logarithm / alpha + sqrt((logarithm / alpha) ** 2 + 4))


def hyperbolic_sinus_prime(u, alpha):
    square = sqrt(4 * (alpha ** 2) * (u ** 2) + 1)
    return (
        0.5 * (((4 * (alpha ** 2) * u) / square + 2 * alpha) / (alpha * (square + 2 * alpha * u)) +
               ((4 * (alpha ** 2) * u / square + 2 * alpha) * ln(square + 2 * alpha * u)) / ((alpha ** 2) * (square + 2 * alpha * u) * sqrt((ln(square + 2 * alpha * u) ** 2) + 4)))
        )


def build_dummies_dict(data):
    """

    Args:
      data:

    Returns:


    """
    unique_val_list = unique(data)
    output = {}
    for val in unique_val_list:
        output[val] = (data == val)
    return output


def calmar(data_in, margins: dict, initial_weight: str, method = 'linear', lo = None, up = None, alpha = None, use_proportions: bool = False,
        xtol: float = 1.49012e-08, maxfev: int = 256):
    """Calibrates weights to satisfy margins constraints.

    Args:
        data_in (pd.DataFrame): The observations data by entity + a dictionary to identify the target entity
        margins (dict): Margins is a dictionnary containing for each variable as key the following values
          - a scalar for numeric variables
          - a dictionnary with categories as key and populations as values
          - eventually a key named `total_population` with value the total population. If absent it is initialized to the actual total population
          - eventually a key named `total_population_smaller_entity` with value the total number of the second entity. If absent it is initialized to the actual total population

        initial_weight (str): Initial weight variable.
        method (str, optional): Calibration method. Should be 'linear', 'raking ratio', 'logit' or 'hyperbolic sinus'. Defaults to 'linear'.
        lo (float, optional): Lower bound on weights ratio. Mandatory when using logit method. Should be < 1. Defaults to None.
        up (float, optional): Upper bound on weights ratio. Mandatory when using logit method. Should be > 1. Defaults to None.
        alpha (float, optional): Bound on weights ratio. Mandatory when using hyperbolic sinus method. Should be > 0. Defaults to None.
        use_proportions (bool, optional): When True use proportions if total population from margins doesn't match total population. Defaults to False.
        xtol (float, optional): Relative precision on lagrangian multipliers.  Defaults to 1.49012e-08 (fsolve xtol).
        maxfev (int, optional): Maximum number of function evaluation. Defaults to 256.

    Raises:
        Exception: [description]
        Exception: [description]
        Exception: [description]

    Returns:
        np.array: Margins adjusting weights
        float: Lagrangian parameter
        dict: Updated margins

    Sources:
        https://github.com/InseeFrLab/Calmar2/blob/main/manuel_utilisation.pdf
    """
    from scipy.optimize import fsolve
    target_entity = data_in['target_entity_name']
    smaller_entity = None
    entities = [target_entity]
    for key in data_in.keys():
        if key != 'target_entity_name' and key != target_entity:
            smaller_entity = key
            entities += [smaller_entity]

    # remove null weights and keep original data
    null_weight_observations = data_in[target_entity][initial_weight].isnull().sum()
    if null_weight_observations > 0:
        log.info("{} observations have a NaN weight. Not used in the calibration.".format(null_weight_observations))

    is_non_zero_weight = (data_in[target_entity][initial_weight].fillna(0) > 0)
    if is_non_zero_weight.sum() > null_weight_observations:
        log.info("{} observations have a zero weight. Not used in the calibration.".format(
            (data_in[target_entity][initial_weight].fillna(0) <= 0).sum() - null_weight_observations))

    variables = set(margins.keys()).intersection(set(data_in[target_entity].columns))
    for variable in variables:
        null_value_observations = data_in[target_entity][variable].isnull().sum()
        if null_value_observations > 0:
            log.info("For variable {}, {} observations have a NaN value. Not used in the calibration.".format(
                variable, null_value_observations))
            is_non_zero_weight = is_non_zero_weight & data_in[target_entity][variable].notnull()

    if not is_non_zero_weight.all():
        log.info("We drop {} observations.".format((~is_non_zero_weight).sum()))

    data = dict()
    if smaller_entity:
        data[smaller_entity] = pd.DataFrame()
        for col in data_in[smaller_entity].columns:
            data[smaller_entity][col] = data_in[smaller_entity][col].copy()
    data[target_entity] = pd.DataFrame()
    for col in data_in[target_entity].columns:
        data[target_entity][col] = data_in[target_entity].loc[is_non_zero_weight, col].copy()

    if not margins:
        raise Exception("Calmar requires non empty dict of margins")

    # choose method
    assert method in ['linear', 'raking ratio', 'logit', 'hyperbolic sinus'], "method should be 'linear', 'raking ratio', 'logit' or 'hyperbolic sinus'"
    if method == 'linear':
        F = linear
        F_prime = linear_prime
    elif method == 'raking ratio':
        F = raking_ratio
        F_prime = raking_ratio_prime
    elif method == 'logit':
        assert up is not None, "When method == 'logit', a value > 1 for up is mandatory"
        assert up > 1, "up should be > 1"
        assert lo is not None, "When method == 'logit', a value < 1 for lo is mandatory"
        assert lo < 1, "lo should be < 1"

        def F(x):
            return logit(x, lo, up)

        def F_prime(x):
            return logit_prime(x, lo, up)
    elif method == 'hyperbolic sinus':
        assert alpha is not None, "When method == 'hyperbolic sinus', a value > 0 for alpha is mandatory"
        assert alpha > 0, "alpha should be > 0"

        def F(x):
            return hyperbolic_sinus(x, alpha)

        def F_prime(x):
            return hyperbolic_sinus_prime(x, alpha)

    margins = margins.copy()
    # Construction observations matrix
    if 'total_population' in margins:
        total_population = margins.pop('total_population')
    else:
        total_population = data[target_entity][initial_weight].fillna(0).sum()
    if smaller_entity is not None:
        if 'total_population_smaller_entity' in margins:
            total_population_smaller_entity = margins.pop('total_population_smaller_entity')
        else:
            total_population_smaller_entity = total_population * len(data[smaller_entity]) / len(data[target_entity])
    else:
        total_population_smaller_entity = 0

    nk = len(data[target_entity][initial_weight])
    # number of Lagrange parameters (at least total population, and potentially total population 2)
    nj = 1 + (smaller_entity is not None)

    margins_new = {}
    margins_new_dict = {}
    for entity in list(entities):
        for var, val in margins.items():
            if var in data[entity].columns:
                if isinstance(val, dict):
                    dummies_dict = build_dummies_dict(data[entity][var])
                    k, pop = 0, 0
                    for cat, nb in val.items():
                        cat_varname = var + '_' + str(cat)
                        data[entity][cat_varname] = dummies_dict[cat]
                        margins_new[cat_varname] = nb
                        if var not in margins_new_dict:
                            margins_new_dict[var] = {}
                        margins_new_dict[var][cat] = nb
                        pop += nb
                        k += 1
                        nj += 1
                    # Check total popualtion
                    population = (entity == target_entity) * total_population + (entity != target_entity) * total_population_smaller_entity
                    if pop != population:
                        if use_proportions:
                            log.info(
                                'calmar: categorical variable {} is inconsistent with population; using proportions'.format(
                                    var
                                    )
                                )
                            for cat, nb in val.items():
                                cat_varname = var + '_' + str(cat)
                                margins_new[cat_varname] = nb * population / pop
                                margins_new_dict[var][cat] = nb * population / pop
                        else:
                            raise Exception('calmar: categorical variable {} weights sums up to {} != {}'.format(
                                var, pop, population))
                else:
                    margins_new[var] = val
                    margins_new_dict[var] = val
                    nj += 1

    # On conserve systematiquement la population
    if hasattr(data, 'dummy_is_in_pop') or hasattr(data, 'dummy_is_in_pop_smaller_entity'):
        raise Exception('dummy_is_in_pop and dummy_is_in_pop_smaller_entity are not valid variable names')

    data[target_entity]['dummy_is_in_pop'] = ones(nk)
    margins_new['dummy_is_in_pop'] = total_population
    if smaller_entity:
        data[smaller_entity]['dummy_is_in_pop_smaller_entity'] = ones(len(data[smaller_entity]['id_variable']))
        margins_new['dummy_is_in_pop_smaller_entity'] = total_population_smaller_entity
    data_final = data[target_entity]

    if smaller_entity:
        liste_col_to_sum = [variable for variable in data[smaller_entity] if variable != "id_variable"]
        dic_agg = {}
        for variable_to_sum in liste_col_to_sum:
            dic_agg[variable_to_sum] = "sum"
        data_second = data[smaller_entity].groupby("id_variable").agg(dic_agg)
        data_final = pd.merge(data_second, data[target_entity], on = "id_variable")
        nk = len(data_final[initial_weight])

    # paramètres de Lagrange initialisés à zéro
    lambda0 = zeros(nj)

    # initial weights
    d = data_final[initial_weight].values
    x = zeros((nk, nj))  # nb obs x nb constraints
    xmargins = zeros(nj)
    margins_dict = {}
    j = 0
    for var, val in margins_new.items():
        x[:, j] = data_final[var]
        xmargins[j] = val
        margins_dict[var] = val
        j += 1

    # Résolution des équations du premier ordre
    def constraint(lambda_):
        return dot(d * F(dot(x, lambda_)), x) - xmargins

    def constraint_prime(lambda_):
        return dot(d * (x.T * F_prime(dot(x, lambda_))), x)
        # le jacobien ci-dessus est constraintprime = @(lambda) x*(d.*Fprime(x'*lambda)*x');

    tries, ier = 0, 2
    err_max = 1
    conv = 1
    while (ier == 2 or ier == 5 or ier == 4) and not (tries >= 10 or (err_max < 1e-6 and conv < 1e-8)):
        lambdasol, infodict, ier, mesg = fsolve(
            constraint,
            lambda0,
            fprime = constraint_prime,
            maxfev = maxfev,
            xtol = xtol,
            full_output = 1,
            )
        lambda0 = 1 * lambdasol
        tries += 1

        pondfin = d * F(dot(x, lambdasol))
        rel_error = {}
        for var, val in margins_new.items():  # noqa analysis:ignore
            rel_error[var] = abs((data_final[var] * pondfin).sum() - margins_dict[var]) / margins_dict[var]
        sorted_err = sorted(rel_error.items(), key = operator.itemgetter(1), reverse = True)

        conv = abs(err_max - sorted_err[0][1])
        err_max = sorted_err[0][1]

    if (ier == 2 or ier == 5 or ier == 4):
        log.debug("optimization converged after {} tries".format(tries))

    # rebuilding a weight vector with the same size of the initial one
    pondfin_out = array(data_in[target_entity][initial_weight], dtype = float64)
    pondfin_out[is_non_zero_weight] = pondfin

    del infodict, mesg  # TODO better exploit this information

    return pondfin_out, lambdasol, margins_new_dict


def check_calmar(margins, margins_new_dict = None):
    """

    Args:
      margins:
      margins_new_dict:  (Default value = None)

    Returns:

    """
    for variable, margin in margins.items():
        if variable != 'total_population':
            print(variable, margin, abs(margin - margins_new_dict[variable]) / abs(margin))  # noqa analysis:ignore



import pandas as pd
from openfisca_country_template.variables.housing import HousingOccupancyStatus
from openfisca_survey_manager.tests import tax_benefit_system

from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario


def test_generation():
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(dict(baseline = tax_benefit_system))
    survey_scenario.period = "2025-06"
    survey_scenario.used_as_input_variables = ['housing_occupancy_status']

    statuses = [HousingOccupancyStatus.free_lodger, HousingOccupancyStatus.tenant]
    data = {
        "input_data_frame": pd.DataFrame({
            "housing_occupancy_status": pd.Series([v.name for v in statuses]),
            "household_id": [0, 1],
            "household_role_index": [0, 0]
            })
        }
    survey_scenario.init_from_data(data=data)
    result = survey_scenario.calculate_variable("housing_occupancy_status", survey_scenario.period)
    assert ((result == pd.Series([v.index for v in statuses])).all())
